from pathlib import Path

import click

from .genall import GenAll


@click.command()
@click.option(
    "--path",
    prompt="Path to the base directory",
    help="The base directory path.",
)
def main(path: str) -> None:
    base_path = Path(path).expanduser().resolve()
    #
    # TODO: path could just be file
    #
    ga = GenAll(base_path)
    ga.write_to_file()



from typing import Optional, overload


def generate_all(items: list[str]) -> str:
    return Assign("__all__", List(items)).build()


def generate_import(from_: str, item: str) -> str:
    return Import(item, from_=f".{from_}").build()


class Buildable:

    def build(self) -> str:
        raise NotImplementedError


class Assign(Buildable):

    def __init__(self, name: str, value: Buildable) -> None:
        self._name = name
        self._value = value

    def build(self) -> str:
        return f"{self._name} = {self._value.build()}"


class List(Buildable):

    def __init__(self, values: list[str]) -> None:
        self._values = values

    def build(self) -> str:
        inner = "".join([f'"{i}", ' for i in self._values])
        return f"[{inner}]"


class Import(Buildable):

    @overload
    def __init__(self, item: str) -> None: ...

    @overload
    def __init__(self, item: str, *items: str, from_: str) -> None: ...

    def __init__(self, item: str, *items: str, from_: Optional[str] = None) -> None:
        self._item = item
        self._items = [item, *items]
        self._from = from_

    def build(self) -> str:
        if self._from is None:
            return f"import {self._item}"

        return f"from {self._from} import {','.join(self._items)}"



from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import Optional

import pathspec
from rich import print

from .codegen import generate_all, generate_import
from .filters.file_filter import ConfigFileFilter
from .parsing import File, PythonObject


def load_gitignore(root: Path) -> pathspec.PathSpec:
    gitignore = root / ".gitignore"

    if not gitignore.exists():
        return pathspec.PathSpec.from_lines("gitwildmatch", [])

    with gitignore.open() as f:
        return pathspec.PathSpec.from_lines("gitwildmatch", f)


def check_ignored(specs: Sequence[pathspec.PathSpec], root: Path, path: Path) -> bool:
    rel_path = path.relative_to(root)

    for spec in reversed(specs):
        if spec.match_file(str(rel_path)):
            return True

    return False


class GenAll:

    def __init__(
        self,
        base_path: Path,
        pathspecs: Optional[list[pathspec.PathSpec]] = None,
    ) -> None:
        self._base_path = base_path
        self._all_objs: list[PythonObject] = []
        self._initialized = False

        if pathspecs is not None:
            self._pathspecs = [
                *pathspecs,
                load_gitignore(base_path),
            ]
        else:
            self._pathspecs = [
                pathspec.PathSpec(
                    [
                        pathspec.RegexPattern(".git"),
                        pathspec.RegexPattern("__init__.py"),
                    ],
                ),
                load_gitignore(base_path),
            ]

    def write_to_file(self) -> None:
        for dir in self._sub_dirs:
            dir.write_to_file()

        if not self._filter_file_path.exists():
            return

        items: list[tuple[str, str]] = []

        for obj in self.all_objs:
            rel = obj._file._path.relative_to(self._base_path)
            parts = rel.parts

            if len(parts) == 1:
                p = rel.stem
            else:
                p = parts[0]

            items.append((p, obj._name))

        imports = [generate_import(*item) for item in items]
        imp = "\n".join(imports)

        code = generate_all([i[1] for i in items])

        file_contents = f"{imp}\n\n{code}"

        with open(self._init_path, "w") as file:
            file.write(file_contents)

        rel = self._init_path.relative_to(self._base_path)
        print(f"[green]+[/green] [bold]Creating file[/bold] [purple]{rel}[/purple]")

    @property
    def all_objs(self) -> list[PythonObject]:
        if not self._initialized:
            self._generate()
            self._initialized = True

        return self._all_objs

    def _generate(self) -> None:
        all_items: list[PythonObject] = []
        output: list[PythonObject] = []

        for file in self._sub_files:
            all_items.extend(file.get_all_objs())

        for dir in self._sub_dirs:
            all_items.extend(dir.all_objs)

        for item in all_items:
            for filter in self._filters:
                if not filter.keep(item):
                    continue

                output.append(item)

        self._all_objs = output

    @property
    def _sub_dirs(self) -> list[GenAll]:
        # TODO: only containing python files
        return [
            GenAll(p, pathspecs=self._pathspecs)
            for p in self._base_path.iterdir()
            if not p.is_file()
            and not check_ignored(self._pathspecs, self._base_path, p)
        ]

    @property
    def _sub_files(self) -> list[File]:
        return [
            File(p)
            for p in self._base_path.iterdir()
            if p.is_file()
            and p.suffix == ".py"
            and not check_ignored(self._pathspecs, self._base_path, p)
        ]

    @property
    def _init_path(self) -> Path:
        return self._base_path / "__init__.py"

    @property
    def _filters(self) -> list[ConfigFileFilter]:
        if not self._filter_file_path.exists():
            return []

        return [ConfigFileFilter.from_file(self._filter_file_path)]

    @property
    def _filter_file_path(self) -> Path:
        return self._base_path / ".genall.yaml"



from __future__ import annotations

import ast
from pathlib import Path


class PythonObject:

    def __init__(self, file: File, type: str, name: str) -> None:
        self._file = file
        self._type = type
        self._name = name

    def __repr__(self) -> str:
        return f"PythonObject({self._type!r}, {self._name!r})"


class File:

    def __init__(self, path: Path) -> None:
        self._path = path

    @property
    def _source(self) -> str:
        with open(self._path, "r") as file:
            return file.read()

    @property
    def _ast(self) -> ast.Module:
        return ast.parse(self._source, self._path.name)

    def get_all_objs(self) -> list[PythonObject]:
        objs: list[PythonObject] = []

        for node in self._ast.body:
            if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                name = node.name
                obj = PythonObject(self, "function", name)
            elif isinstance(node, ast.ClassDef):
                name = node.name
                obj = PythonObject(self, "class", name)
            elif isinstance(node, ast.Assign):
                # TODO: multiple assignments...
                name = node.targets[0].id  # type: ignore
                obj = PythonObject(self, "variable", name)
            else:
                continue

            if name.startswith("_"):
                continue

            objs.append(obj)

        return sorted(objs, key=lambda x: x._name)



from __future__ import annotations

import re
from pathlib import Path
from typing import Optional

from yaml import safe_load

from genall.parsing import PythonObject


class ConfigFileFilter:

    def __init__(
        self,
        classes: Optional[list[Filter]] = None,
        functions: Optional[list[Filter]] = None,
        variables: Optional[list[Filter]] = None,
    ) -> None:
        self._classes = classes
        self._functions = functions
        self._variables = variables

    @classmethod
    def from_file(cls, path: Path) -> ConfigFileFilter:
        with open(path, "r") as f:
            data = safe_load(f)

        if data is None:
            data = {}

        classes: list[Filter] | None = None
        functions: list[Filter] | None = None
        variables: list[Filter] | None = None

        if data.get("classes") is not None:
            classes = []
            for item in data["classes"]:
                filter = Filter(
                    type="class",
                    name=item["name"],
                    include=item.get("include", True),
                )
                classes.append(filter)

        if data.get("functions") is not None:
            functions = []
            for item in data["functions"]:
                filter = Filter(
                    type="function",
                    name=item["name"],
                    include=item.get("include", True),
                )
            functions.append(filter)

        if data.get("variables") is not None:
            variables = []
            for item in data["variables"]:
                filter = Filter(
                    type="variable",
                    name=item["name"],
                    include=item.get("include", True),
                )
            variables.append(filter)

        return cls(
            classes=classes,
            functions=functions,
            variables=variables,
        )

    def keep(self, obj: PythonObject) -> bool:
        return self._keep_by_type(obj)

    def _keep_obj_filters(
        self,
        obj: PythonObject,
        filters: Optional[list[Filter]] = None,
    ) -> bool:
        if filters is None:
            return True

        for filter in filters:
            if not filter.matches(obj):
                if not filter._include:
                    return True

                continue

            return filter.keep(obj)

        return False

    def _keep_class(self, obj: PythonObject) -> bool:
        return self._keep_obj_filters(obj, self._classes)

    def _keep_function(self, obj: PythonObject) -> bool:
        return self._keep_obj_filters(obj, self._functions)

    def _keep_variable(self, obj: PythonObject) -> bool:
        return self._keep_obj_filters(obj, self._variables)

    def _keep_by_type(self, obj: PythonObject) -> bool:
        if obj._type == "class":
            return self._keep_class(obj)
        elif obj._type == "function":
            return self._keep_function(obj)
        elif obj._type == "variable":
            return self._keep_variable(obj)

        return True


class Filter:

    def __init__(self, type: str, name: str, include: bool) -> None:
        self._type = type
        self._name = name
        self._include = include

    def matches(self, obj: PythonObject) -> bool:
        if obj._type != self._type:
            return False

        if not self._name.startswith("/") or not self._name.endswith("/"):
            return obj._name == self._name

        pattern = self._name[1:-1]
        return re.match(pattern, obj._name) is not None

    def keep(self, obj: PythonObject) -> bool:
        if not self.matches(obj):
            return False

        return self._include

    def __repr__(self) -> str:
        return f"Filter(type={self._type!r}, name={self._name!r}, include={self._include!r})"
